@@ -1,26 +1,26 @@
-from sqlalchemy import create_engine, inspect, func, MetaData
-from sqlalchemy.orm import sessionmaker, Session, scoped_session
-from sqlalchemy.exc import (
-    ArgumentError,
-    InvalidRequestError,
-    OperationalError,
-    IntegrityError,
-)
-from sqlalchemy.ext.automap import automap_base
-from sqlalchemy.engine.url import make_url
-from sqlalchemy.ext.declarative.api import DeclarativeMeta
-from sqlalchemy.pool import NullPool
-from marshmallow import EXCLUDE
-from emannotationschemas import get_schema, get_flat_schema
-from emannotationschemas.flatten import flatten_dict
+import datetime
+import logging
+from typing import List
+
+from emannotationschemas import get_schema
 from emannotationschemas import models as em_models
+from emannotationschemas.flatten import flatten_dict
+from marshmallow import EXCLUDE
+import marshmallow
+from sqlalchemy import create_engine, func, inspect, DDL, event
+from sqlalchemy.engine.url import make_url
+from sqlalchemy.ext.automap import automap_base
+from sqlalchemy.ext.declarative.api import DeclarativeMeta
+from sqlalchemy.orm import Session, scoped_session, sessionmaker
+from sqlalchemy.pool import NullPool
+
+from dynamicannotationdb.errors import (
+    TableAlreadyExists,
+    TableNameNotFound,
+    SelfReferenceTableError,
+)
 from dynamicannotationdb.key_utils import build_segmentation_table_name
 from dynamicannotationdb.models import AnnoMetadata, SegmentationMetadata
-from dynamicannotationdb.errors import TableNameNotFound, TableAlreadyExists
-from typing import List
-import logging
-import datetime
-import time
 
 
 class DynamicAnnotationInterface:
@@ -102,13 +102,13 @@ class DynamicAnnotationInterface:
 
                 # check if template exists, create if missing
                 template_exist = connection.execute(
-                    f"SELECT 1 FROM pg_catalog.pg_database WHERE datname = 'template_postgis'"
+                    "SELECT 1 FROM pg_catalog.pg_database WHERE datname = 'template_postgis'"
                 )
 
                 if not template_exist.fetchone():
 
                     # create postgis template db
-                    connection.execute(f"CREATE DATABASE template_postgis")
+                    connection.execute("CREATE DATABASE template_postgis")
 
                     # create postgis extension
                     template_uri = make_url(f"{sql_base_uri}/template_postgis")
@@ -141,13 +141,22 @@ class DynamicAnnotationInterface:
     def commit_session(self):
         try:
             self.cached_session.commit()
-        except Exception:
+        except Exception as e:
             self.cached_session.rollback()
-            logging.exception(f"SQL Error")
+            logging.exception(f"SQL Error: {e}")
+            raise (e)
         finally:
             self.cached_session.close()
         self._cached_session = None
 
+    def check_table_is_unique(self, table_name):
+        existing_tables = self._get_existing_table_names()
+        if table_name in existing_tables:
+            raise TableAlreadyExists(
+                f"Table creation failed: {table_name} already exists"
+            )
+        return existing_tables
+
     def create_annotation_table(
         self,
         table_name: str,
@@ -157,8 +166,9 @@ class DynamicAnnotationInterface:
         voxel_resolution_x: float,
         voxel_resolution_y: float,
         voxel_resolution_z: float,
-        reference_table: str = None,
+        table_metadata: dict = None,
         flat_segmentation_source: str = None,
+        with_crud_columns: bool = True,
     ):
         r"""Create new annotation table unless already exists
 
@@ -171,7 +181,7 @@ class DynamicAnnotationInterface:
         schema_type : str
             Type of schema to use, must be a valid type from EMAnnotationSchemas
 
-         description: str
+        description: str
             a string with a human readable explanation of
             what is in the table. Including who made it
             and any information that helps interpret the fields
@@ -189,19 +199,43 @@ class DynamicAnnotationInterface:
         voxel_resolution_z: float
             voxel_resolution of this annotation table's point in z (typically nm)
 
-        reference_table: str
-            reference table name, if required by this schema
+        table_metadata: dict
 
         flat_segmentation_source: str
             a path to a segmentation source associated with this table
              i.e. 'precomputed:\\gs:\\my_synapse_seg\example1'
+
+        with_crud_columns: bool
+            add additional columns to track CRUD operations on rows
         """
+        existing_tables = self.check_table_is_unique(table_name)
+        reference_table = None
+        track_updates = None
 
-        if table_name in self._get_existing_table_names():
-            logging.warning(f"Table creation failed: {table_name} already exists")
-            raise TableAlreadyExists
+        if table_metadata:
+            try:
+                reference_table = table_metadata.get("reference_table")
+            except AttributeError as e:
+                reference_table = None
+            try:
+                track_updates = table_metadata.get("track_target_id_updates")
+            except AttributeError as e:
+                track_updates = None
 
-        model = em_models.make_annotation_model(table_name, schema_type)
+            if reference_table:
+                reference_table = self._parse_reference_table_metadata(
+                    table_name, table_metadata, existing_tables
+                )
+                self.base.metadata.reflect()
+
+        model = em_models.make_annotation_model(
+            table_name, schema_type, table_metadata, with_crud_columns
+        )
+
+        if reference_table and track_updates:
+            description = self.create_reference_update_trigger(
+                table_name, description, reference_table, model
+            )
 
         self.base.metadata.tables[model.__name__].create(bind=self.engine)
         creation_time = datetime.datetime.now()
@@ -260,7 +294,6 @@ class DynamicAnnotationInterface:
         model = em_models.make_segmentation_model(
             annotation_table_name, schema_type, pcg_table_name
         )
-
         self.base.metadata.tables[model.__name__].create(bind=self.engine)
         creation_time = datetime.datetime.now()
 
@@ -285,13 +318,13 @@ class DynamicAnnotationInterface:
         metadata = (
             self.cached_session.query(AnnoMetadata)
             .filter(AnnoMetadata.table_name == table_name)
-            .first()
+            .one()
         )
-        try:
-            metadata.__dict__.pop("_sa_instance_state")
-            return metadata.__dict__
-        except Exception as e:
-            raise AttributeError(f"No table found with name '{table_name}'. Error: {e}")
+        if not metadata:
+            raise TableNameNotFound(
+                f"Error: No table name exists with name {table_name}."
+            )
+        return metadata.__dict__
 
     def get_segmentation_table_metadata(self, table_name: str, pcg_table_name: str):
         seg_table_name = build_segmentation_table_name(table_name, pcg_table_name)
@@ -300,11 +333,48 @@ class DynamicAnnotationInterface:
             .filter(SegmentationMetadata.table_name == seg_table_name)
             .first()
         )
-        try:
-            metadata.__dict__.pop("_sa_instance_state")
-            return metadata.__dict__
-        except Exception as e:
-            raise AttributeError(f"No table found with name '{table_name}'. Error: {e}")
+        if not metadata:
+            raise TableNameNotFound(
+                f"Error: No table name exists with name {table_name}."
+            )
+        return metadata.__dict__
+
+    def create_reference_update_trigger(
+        self, table_name, description, reference_table, model
+    ):
+        func_name = f"{table_name}_update_reference_id"
+        func = DDL(
+            f"""
+                    CREATE or REPLACE function {func_name}()
+                    returns TRIGGER
+                    as $func$
+                    begin
+                        update {table_name} ref
+                        set target_id = new.superceded_id
+                        where ref.target_id = old.id;
+                        return new;
+                    end;
+                    $func$ language plpgsql;
+                    """
+        )
+        trigger = DDL(
+            f"""CREATE TRIGGER update_{table_name}_target_id AFTER UPDATE ON {reference_table}
+                    FOR EACH ROW EXECUTE PROCEDURE {func_name}();"""
+        )
+
+        event.listen(
+            model.__table__,
+            "after_create",
+            func.execute_if(dialect="postgresql"),
+        )
+
+        event.listen(
+            model.__table__,
+            "after_create",
+            trigger.execute_if(dialect="postgresql"),
+        )
+        description += f" [Note: This table '{model.__name__}' will update the 'target_id' foreign_key when updates are made to the '{reference_table}' table]"
+        return description
 
     def _get_existing_table_by_name(self) -> List[str]:
         """Get the table names of table that exist
@@ -330,15 +400,15 @@ class DynamicAnnotationInterface:
             for table_name in self._get_existing_table_names()
         ]
 
-    def get_table_schema(self, table_name: str):
+    def get_table_schema(self, table_name: str) -> str:
         table_metadata = self.get_table_metadata(table_name)
         return table_metadata["schema_type"]
 
-    def get_table_sql_metadata(self, table_name):
+    def get_table_sql_metadata(self, table_name: str):
         self.base.metadata.reflect(bind=self.engine)
         return self.base.metadata.tables[table_name]
 
-    def _get_model_columns(self, table_name: str) -> list:
+    def _get_model_columns(self, table_name: str) -> List[tuple]:
         """Return list of column names and types of a given table
 
         Parameters
@@ -351,21 +421,17 @@ class DynamicAnnotationInterface:
         list
             column names and types
         """
-        try:
-            model_columns = []
-            db_columns = self.insp.get_columns(table_name)
-            for column in db_columns:
-                model_columns.append(tuple([column["name"], column["type"]]))
-            return model_columns
-        except Exception as e:
+        db_columns = self.insp.get_columns(table_name)
+        if not db_columns:
             raise TableNameNotFound(
-                f"Error: {e}. No table name exists with name {table_name}."
+                f"Error: No table name exists with name {table_name}."
             )
+        return [tuple([column["name"], column["type"]]) for column in db_columns]
 
     def _get_all_tables(self):
         return self.cached_session.query(AnnoMetadata).all()
 
-    def _get_existing_table_names(self):
+    def _get_existing_table_names(self) -> List[str]:
         """Collects table_names keys of existing tables
 
         Returns
@@ -376,11 +442,11 @@ class DynamicAnnotationInterface:
         metadata = self.cached_session.query(AnnoMetadata).all()
         return [m.table_name for m in metadata]
 
-    def get_valid_table_names(self):
+    def get_valid_table_names(self) -> List[str]:
         metadata = self.cached_session.query(AnnoMetadata).all()
         return [m.table_name for m in metadata if m.valid == True]
 
-    def get_existing_segmentation_table_names(self):
+    def get_existing_segmentation_table_names(self) -> List[str]:
         """Collects table_names keys of existing segmentation tables
         contained in an aligned volume database. Used for materialization.
 
@@ -392,6 +458,14 @@ class DynamicAnnotationInterface:
         metadata = self.cached_session.query(SegmentationMetadata).all()
         return [m.table_name for m in metadata]
 
+    def get_reference_annotation_table_names(self, target_table: str) -> List[str]:
+        metadata = (
+            self.cached_session.query(AnnoMetadata)
+            .filter(AnnoMetadata.reference_table == target_table)
+            .all()
+        )
+        return [table for table in metadata]
+
     def has_table(self, table_name: str) -> bool:
         """Check if a table exists
 
@@ -461,7 +535,7 @@ class DynamicAnnotationInterface:
         Model = self._cached_table(table_name)
         return self.cached_session.query(Model).count()
 
-    def _drop_table(self, table_name: str):
+    def _drop_table(self, table_name: str) -> bool:
         table = self.base.metadata.tables.get(table_name)
         if table:
             logging.info(f"Deleting {table_name} table")
@@ -497,9 +571,10 @@ class DynamicAnnotationInterface:
             flat_segmentation_schema,
         ) = em_models.split_annotation_schema(schema_type)
 
-        return self._map_values_to_schema(
-            data, flat_annotation_schema
-        ), self._map_values_to_schema(data, flat_segmentation_schema)
+        return (
+            self._map_values_to_schema(data, flat_annotation_schema),
+            self._map_values_to_schema(data, flat_segmentation_schema),
+        )
 
     def _get_flattened_schema(self, schema_type: str):
         schema_type = get_schema(schema_type)
@@ -511,7 +586,7 @@ class DynamicAnnotationInterface:
 
         return flat_annotation_schema, flat_segmentation_schema
 
-    def _map_values_to_schema(self, data, schema):
+    def _map_values_to_schema(self, data: dict, schema: marshmallow.Schema):
         return {
             key: data[key]
             for key, value in schema._declared_fields.items()
@@ -534,6 +609,22 @@ class DynamicAnnotationInterface:
 
         return table_name in self._cached_tables
 
+    def _parse_reference_table_metadata(
+        self, table_name, table_metadata, existing_tables
+    ):
+        reference_table = table_metadata.get("reference_table")
+
+        if table_name is reference_table:
+            raise SelfReferenceTableError(
+                f"{reference_table} must target a different table not {table_name}"
+            )
+        if reference_table not in existing_tables:
+            raise TableNameNotFound(
+                f"Reference table target: '{existing_tables}' does not exist"
+            )
+
+        return reference_table
+
     def _load_table(self, table_name: str):
         """Load existing table into cached lookup dict instance
 
